// util.c

/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

// #include <conio.h>	// gb - this is not C, was used for getche()
#include <stdio.h>
#include <stdarg.h>
//#include <malloc.h>
#include <memory.h>
#include "qbsp.h"

extern char *rgszWarnings[cWarnings];
extern char *rgszErrors[cErrors];

int rgMemActive[GLOBAL+1];
int rgMemPeak[GLOBAL+1];
int cTotalWarnings = 0;

/*
==========
AllocMem
==========
*/
void *AllocMem (int Type, int cElements, bool fZero)
{
	void *pTemp;
	int  cSize, cWindPoints;

	if (Type < 0 || Type > OTHER)
		Message(msgError, errInvalidMemType, Type);

	// For windings, cElements == number of points on winding
	if (Type == WINDING)
	{
		if (cElements > MAX_POINTS_ON_WINDING)
			Message(msgError, errTooManyPoints, cElements, MAX_POINTS_ON_WINDING);

		cSize = (int)((winding_t *)0)->points[cElements];

		cWindPoints = cElements;
	}
	else
		cSize = cElements * rgcMemSize[Type];

	if (Type == WINDING || Type == OTHER)
	{
		cSize += sizeof(int); // Extra storage for size
	
		// Adjust cElements so bookkeeping works OK
		if (Type == WINDING)
			cElements = 1;
		else
			cElements = cSize;
	}

Retry:
	pTemp = (void *)malloc(cSize);

	if (pTemp == NULL)
	{
		char ch;

		printf("\bOut of memory.  Please close some apps and hit 'y' to try again,\n");
		printf("or any other key to exit -> ");
		ch = toupper(getchar());	// gb, was getche() 
		printf("\n");
		fflush(stdin);
		if (ch == 'Y')
			goto Retry;
		Message(msgError, errOutOfMemory);
	}

	if (fZero)
		memset(pTemp, 0, cSize);

	if (Type == WINDING || Type == OTHER)
	{
		// Save total size just before the user area
		*(int *)pTemp = cSize;
		pTemp = (char *)pTemp + sizeof(int);
	}

	// Special stuff
	if (Type == FACE)
		((face_t *)pTemp)->planenum = -1;
	else if (Type == WINDING)
		((winding_t *)pTemp)->numpoints = cWindPoints;

	rgMemActive[Type] += cElements;
	if (rgMemActive[Type] > rgMemPeak[Type])
		rgMemPeak[Type] = rgMemActive[Type];

	// Also keep global statistics
	rgMemActive[GLOBAL] += cSize;
	if (rgMemActive[GLOBAL] > rgMemPeak[GLOBAL])
		rgMemPeak[GLOBAL] = rgMemActive[GLOBAL];

	return pTemp;
}


/*
==========
FreeMem
==========
*/
void FreeMem (void *pMem, int Type, int cElements)
{
	if (Type == OTHER)
		cElements += sizeof(int); // Adjust for extra size

	rgMemActive[Type] -= cElements;

	if (Type == FACE)
		ResizeFace ((face_t *)pMem, 0);

	if (Type == WINDING || Type == OTHER)
	{
		// Total size is saved just before the user area
		pMem = (char *)pMem - sizeof(int);
		rgMemActive[GLOBAL] -= *(int *)pMem;
	}
	else
	        rgMemActive[GLOBAL] -= cElements * rgcMemSize[Type];

	free(pMem);
}

void ResizeFace (face_t *f, int numpoints)
{
	// Note: f->numpoints may not be correct here
	f->pts = (vec3_t *)ReAllocOther (f->pts, numpoints * sizeof(vec3_t));
	f->numpoints = numpoints;
}

void CopyFace (face_t *dst, face_t *src)
{
	vec3_t *Save;

	ResizeFace (dst, src->numpoints);

	Save = dst->pts;
	*dst = *src;
	dst->pts = Save;

	memcpy (dst->pts, src->pts, dst->numpoints * sizeof(vec3_t));
}

void *ReAllocOther (void *o, size_t Size)
{
	void *o2;
	int  OldSize;

	o2 = Size == 0 ? NULL : AllocMem (OTHER, Size);

	if (o != NULL)
	{
		// Total size is saved just before the user area
		OldSize = *(int *)((char *)o - sizeof(int)) - sizeof(int);

		if (o2 != NULL)
			memcpy (o2, o, (int)Size < OldSize ? Size : OldSize);

		FreeMem (o, OTHER, OldSize);
	}

	return(o2);
}


/*
==========
PrintMem
==========
*/
void PrintMem(void)
{
	char *rgszMemTypes[] = {"BSPEntity", "BSPPlane", "BSPTex", "BSPVertex", "BSPVis",
		"BSPNode", "BSPTexinfo", "BSPFace", "BSPLight", "BSPClipnode", "BSPLeaf",
		"BSPMarksurface", "BSPEdge", "BSPSurfedge", "BSPModel", "Mapface", "Mapbrush",
		"Mapentity","Winding", "Face", "Plane", "Portal", "Surface", "Node", "Brush",
		"Miptex", "World verts", "World edges", "Hash verts", "Other (kbyte)",
		"Total (kbyte)"};
	int i, Factor;

	if (options.fAllverbose || options.fMemverbose)
	{
		Message(msgFile, "\nData type     Current     Peak  kbyte\n");
		for (i=0; i<=GLOBAL; i++)
		{
			Factor = i < GLOBAL - 1 ? 1 : 1024;

			Message(msgFile, "%-14s%7d %8d %6d\n", rgszMemTypes[i], rgMemActive[i] / Factor,
				rgMemPeak[i] / Factor, rgMemPeak[i] * rgcMemSize[i] / 1024);
		}
	}

	Message(msgLiteral, "\nPeak memory used : %.1f MB\n", (float)rgMemPeak[GLOBAL] / (1024 * 1024));
}

/*
============
FreeAllMem
============
*/
void FreeAllMem(void)
{
	int i, j;
	epair_t *ep, *next;

	for (i=0; i<map.cEntities; i++)
	{
		for (ep = map.rgEntities[i].epairs; ep; ep=next)
		{
			next = ep->next;
			if (ep->key)
				FreeMem(ep->key, OTHER, strlen(ep->key)+1);
			if (ep->value)
				FreeMem(ep->value, OTHER, strlen(ep->value)+1);
			FreeMem(ep, OTHER, sizeof(epair_t));
		}
		for (j=0; j<BSP_LUMPS; j++)
			if (map.rgEntities[i].pData[j])
				FreeMem(map.rgEntities[i].pData[j], j, map.rgEntities[i].cData[j]);
	}

	FreeMem(validfaces, OTHER, sizeof(face_t *)*cPlanes);
	FreeMem(pPlanes, PLANE, cPlanes);
	FreeMem(map.rgFaces, MAPFACE, map.cFaces);
	FreeMem(map.rgBrushes, MAPBRUSH, map.cBrushes);
	FreeMem(map.rgEntities, MAPENTITY, map.cEntities);

}


/*
=================
Message

Generic output of errors, warnings, stats, etc
=================
*/
void Message(int msgType, ...)
{
	if( msgType == msgProgress || msgType == msgPercent )
	{
		return;
	}
	
	
	va_list	    argptr;
	char	    szBuffer[512];
	char	    *szFmt;
	int	    ErrType;
	int	    Cur, Total;
	bool	    fDisableScreen = false;
	static bool fInPercent = false;

	va_start(argptr, msgType);

	// Exit or disable screen output as necessary
	if ((msgType == msgStat || msgType == msgStat2 || msgType == msgProgress) && (!options.fVerbose || options.fNoverbose))
	{
		if (!options.fVerboseFile)
			return;
		
		fDisableScreen = true;
	}
	else if (msgType == msgPercent && options.fNopercent)
		return;

	// Grab proper args
	if (msgType == msgWarning || msgType == msgError)
	{
		ErrType = va_arg(argptr, int);
		
		if (msgType == msgWarning)
		{
			++cTotalWarnings;

			// Disable screen output except for critical warnings
			if (options.fNowarnings && ErrType != warnNoValidWads &&
						   ErrType != warnNoPlayerStart &&
						   ErrType != warnMapLeak &&
						   ErrType != warnDiffModels &&
						   ErrType != warnManyVisleafs &&
						   ErrType != warnNoFilling &&
						   ErrType != warnManyClipnodes &&
						   ErrType != warnManyMarksurfaces &&
						   ErrType != warnManyModels &&
						   ErrType != warnManyPlanes &&
						   ErrType != warnManyNodes &&
						   ErrType != warnManyLeafs &&
						   ErrType != warnManyVertexes &&
						   ErrType != warnManyFaces &&
						   ErrType != warnManyTexinfo)
				fDisableScreen = true;
		}
	}
	else if (msgType == msgPercent)
	{
		Cur = va_arg(argptr, int);
		Total = va_arg(argptr, int);

		if (Total == -1)
		{
			fInPercent = false;
			ShowBar(-1, -1);
			return;
		}
	}
	else
		szFmt = va_arg(argptr, char *);

	// Handle warning/error-in-percent properly
	if (fInPercent && msgType != msgPercent && !fDisableScreen)
		ShowBar(0, -1);

	switch (msgType)
	{
	case msgWarning:
		if (ErrType >= cWarnings)
		{
			printf("Internal error: Unknown ErrType (%d) in Message (msgWarning) !\n", ErrType);
			LogFile.Close();
			exit(1);
		}
		sprintf(szBuffer, "*** WARNING %02d: ", ErrType);
		vsprintf(szBuffer+strlen(szBuffer), rgszWarnings[ErrType], argptr);
		strcat(szBuffer, "\n");
		break;

	case msgError:
		if (ErrType >= cErrors)
			printf("Internal error: Unknown ErrType (%d) in Message (msgError) !\n", ErrType);
		else
		{
			sprintf(szBuffer, "*** ERROR %02d: ", ErrType);
			vsprintf(szBuffer+strlen(szBuffer), rgszErrors[ErrType], argptr);
			puts(szBuffer);
			LogFile.Printf("%s\n", szBuffer);
		}
		PrintFinish();
		LogFile.Close();
		exit(1);
		break;

	case msgLiteral:
		// Output as-is to screen and log file
		vsprintf(szBuffer, szFmt, argptr);
		break;

	case msgStat:
	case msgStat2:
		// Output as-is to screen and log file
		szBuffer[0] = '\0';
		if (msgType == msgStat)
			strcpy(szBuffer, "       ");

		vsprintf(szBuffer+strlen(szBuffer), szFmt, argptr);		// Concatenate
		strcat(szBuffer, "\n");
		break;

	case msgProgress:
		// Output as-is to screen and log file
		strcpy(szBuffer, "---- ");
		vsprintf(szBuffer+strlen(szBuffer), szFmt, argptr);		// Concatenate
		strcat(szBuffer, " ----\n");
		break;

	case msgPercent:
		// Calculate the percent complete.  Only output if it changes.
		if (!fInPercent)
			ShowBar(0, 0);

		ShowBar(Cur, Total);

		// Handle output formatting properly
		fInPercent = true;
		return;

	case msgFile:
		// Output only to the file
		vsprintf(szBuffer, szFmt, argptr);
		break;

	case msgScreen:
		// Output only to the screen
		vsprintf(szBuffer, szFmt, argptr);
		break;

	default:
		printf("Unhandled msgType in message!\n");
		return;
	}

	if (msgType != msgFile && !fDisableScreen)
	{
		printf("%s", szBuffer);
		fflush(stdout);
	}

	if (msgType != msgScreen)
	{
		LogFile.Printf("%s", szBuffer);
		LogFile.Flush();
	}

	va_end(argptr);
}

/*
===============
szGetCoord

Returns coordinates as a string
===============
*/
char *szGetCoord(vec3_t v)
{
	static char szStr[50];

	sprintf (szStr, "(%.0f %.0f %.0f)", v[0], v[1], v[2]);

	return szStr;
}
